
#include "RdmaEndpoint.hpp"
#include "Logger.hpp"
int RdmaEndpoint::CONN_STATE_INITIALIZED = 1;
int RdmaEndpoint::CONN_STATE_RESOURCES_ALLOCATED = 2;
int RdmaEndpoint::CONN_STATE_CONNECTED = 3;
int RdmaEndpoint::CONN_STATE_CLOSED = 4;

RdmaEndpoint::RdmaEndpoint(struct rdma_cm_id *id, struct ibv_cq *completionQueue, int sendQueueSize,
                           int recvQueueSize, int sendMsgSize, int recvMsgSize)
    : _cm_id(id), _completionQueue(completionQueue), _sendQueueSize(sendQueueSize),
      _recvQueueSize(recvQueueSize), _sendMsgSize(sendMsgSize), _recvMsgSize(recvMsgSize)
{
    _state = CONN_STATE_INITIALIZED;
}
void RdmaEndpoint::createResources()
{
    /* These states are used to avoid errors in lifetime of rdma connection
     * more erros can be tracked in future using these lifecycle states
     */
    if (_state != CONN_STATE_INITIALIZED)
    {
        CPPLog::LOG_ERROR("RdmaEndpoint : createResource invalid state");
    }
    //Step 1 to create endpoint
    _protectionDomain = ibv_alloc_pd(_cm_id->verbs);
    if (_protectionDomain == NULL)
    {
        CPPLog::LOG_ERROR("RdmaEndpoint : ibv_alloc_pd failed");
        return;
    }
    //step 2 Creating Queue pair with completion queueu setted for send and recieve
    struct ibv_qp_init_attr qp_init_attr;
    memset(&qp_init_attr, 0, sizeof(qp_init_attr));
    /*
     *  Endpoint address is setted in QP context to get endpoint at run time with qp
     *  without using any map to map qp_num to endpoint
     */
    qp_init_attr.qp_context = (void *)this;
    // if not set 0, all work requests submitted to SQ will always generate a Work Completion
    qp_init_attr.sq_sig_all = 1;
    // completion queue can be shared or you can use distinct completion queues.
    qp_init_attr.send_cq = _completionQueue;
    qp_init_attr.recv_cq = _completionQueue;
    qp_init_attr.qp_type = IBV_QPT_RC;
    // increase if you want to keep more send work requests in the SQ.
    qp_init_attr.cap.max_send_wr = _sendQueueSize;
    // increase if you want to keep more receive work requests in the RQ.
    qp_init_attr.cap.max_recv_wr = _recvQueueSize;
    // increase if you allow send work requests to have multiple scatter gather entry (SGE).
    qp_init_attr.cap.max_send_sge = 1;
    // increase if you allow receive work requests to have multiple scatter gather entry (SGE).
    qp_init_attr.cap.max_recv_sge = 1;

    int ret = rdma_create_qp(_cm_id, _protectionDomain, &qp_init_attr);
    if (ret)
    {
        CPPLog::LOG_ERROR("RdmaEndpoint : ibv_create_cq failed");
    }
    if (_cm_id->pd == NULL)
    {
        CPPLog::LOG_ERROR("RdmaEndpoint : pd not set");
        _cm_id->pd = _protectionDomain;
    }
    /*
    * Step 3 register memory for send and recv queue
    */
    _sendBuff = new char[(_sendMsgSize * _sendQueueSize)];
    if (_sendBuff == NULL)
        CPPLog::LOG_ERROR("RdmaEndpoint :  sendBuff malloc failed");

    _sendMr = rdma_reg_write(_cm_id, reinterpret_cast<void *>(_sendBuff), _sendMsgSize * _sendQueueSize);
    if (_sendMr == NULL)
        CPPLog::LOG_ERROR("RdmaEndpoint :  sendMr reg failed");
    
    _recvBuff = new char[(_recvMsgSize * _recvQueueSize)];
    if (_recvBuff == NULL)
        CPPLog::LOG_ERROR("RdmaEndpoint :  recvBuff malloc failed");
    _recvMr = rdma_reg_read(_cm_id, reinterpret_cast<void *>(_recvBuff), _recvMsgSize * _recvQueueSize);

    if (_recvMr == NULL)
        CPPLog::LOG_ERROR("RdmaEndpoint :  recvMr reg failed");

    /*
    * adding buffers for recving rdma data
    */
    for (int i = 0; i < _recvQueueSize; i++)
    {
        char* location = _recvBuff + i * _recvMsgSize;
        rdma_post_recv(_cm_id, reinterpret_cast<void *>(location), reinterpret_cast<void *>(location),
                       _recvMsgSize, _recvMr);
    }
    /*
    *  Adding buffers to queue for receving data
    */
    for (int i = 0; i < _sendQueueSize; i++)
    {
        char* location = _sendBuff + i * _sendMsgSize;
        std::unique_lock<std::mutex> lock(_sendBuffersM);
        _sendBuffers.push((char *)location);
    }
    _state = CONN_STATE_RESOURCES_ALLOCATED;
}

void RdmaEndpoint::processCmEvent(struct rdma_cm_event *event)
{
    std::ostringstream ss;
    ss<<"RdmaEndpoint : Event "<<rdma_event_str(event->event);
    CPPLog::LOG_ALWAYS(ss);
    if (event->event == RDMA_CM_EVENT_CONNECT_REQUEST)
    {
        CPPLog::LOG_ALWAYS("RdmaEndpoint : Connect request");
    }
    else if (event->event == RDMA_CM_EVENT_ESTABLISHED)
    {
        if (_state != CONN_STATE_RESOURCES_ALLOCATED)
        {
            CPPLog::LOG_ERROR("RdmaEndpoint : Established_Event but resource not alloted");
        }
        CPPLog::LOG_INFO("RdmaEndpoint : step 6 Connected");
        _state = CONN_STATE_CONNECTED;
    }
    else if (event->event == RDMA_CM_EVENT_DISCONNECTED)
    {
        CPPLog::LOG_INFO("RdmaEndpoint : step 7 disconnected");
        close();
    }
}

void RdmaEndpoint::close()
{
    if (_state < CONN_STATE_RESOURCES_ALLOCATED)
    {
        return;
    }
    CPPLog::LOG_INFO("RdmaEndpoint : closing connection");
    int ret;
    ret = rdma_disconnect(_cm_id);
    if (ret)
    {
        CPPLog::LOG_ERROR("RdmaEndpoint : rdma_disconnect failed");
    }

    ret = rdma_dereg_mr(_sendMr);
    if (ret)
    {
        CPPLog::LOG_ERROR("RdmaEndpoint : rdma_dereg_mr send failed");
    }
    delete[] _sendBuff;

    ret = rdma_dereg_mr(_recvMr);
    if (ret)
    {
        CPPLog::LOG_ERROR("RdmaEndpoint : rdma_dereg_mr recv failed");
    }
    delete[] _recvBuff;

    rdma_destroy_qp(_cm_id);

    CPPLog::LOG_INFO("des qp");
    // rdma_destroy_id(_cm_id);
    // ret = rdma_destroy_id(_cm_id);

    CPPLog::LOG_INFO("des mr");
    if (ret)
    {
        CPPLog::LOG_ERROR("RdmaEndpoint : rdma_destroy_id failed");
    }
    _state = CONN_STATE_CLOSED;
    CPPLog::LOG_INFO("closed");
}

RdmaEndpoint::~RdmaEndpoint()
{}