#include "RdmaSalEndpoint.hpp"
#include "MessageFormats.hpp"
RdmaSalEndpoint::RdmaSalEndpoint(struct rdma_cm_id *id, struct ibv_cq *completionQueue, int sendQueueSize,
                                 int recvQueueSize, int sendMsgSize, int recvMsgSize, rocksdb::DB *db)
    : RdmaEndpoint(id, completionQueue, sendQueueSize, recvQueueSize, sendMsgSize, recvMsgSize), _db(db)
{
}

void RdmaSalEndpoint::processSendCompletion(struct ibv_wc *data)
{
    /*means data has been send to other side we can use this buffer*/
    std::unique_lock<std::mutex> lock(_sendBuffersM);
    _sendBuffers.push((char *)data->wr_id);
}

int RdmaSalEndpoint::sendMessage(const char *buffer, uint32_t size)
{
    if (size > (uint32_t)_sendMsgSize)
        return -1;
    char *sendBuffer = nullptr;
    std::unique_lock<std::mutex> lock(_sendBuffersM);
    if (_sendBuffers.size() == 0)
        return -1;
    sendBuffer = _sendBuffers.front();
    _sendBuffers.pop();
    lock.unlock();
    memcpy((void *)sendBuffer, buffer, size);
    return rdma_post_send(_cm_id, (void *)sendBuffer, (void *)sendBuffer, size, _sendMr, 0);
}

void RdmaSalEndpoint::processRecvCompletion(struct ibv_wc *data)
{
    char *request = new char[data->byte_len];
    memcpy(request, (void *)data->wr_id, data->byte_len);
    struct MessageHeader *req = (struct MessageHeader *)request;
    rdma_post_recv(_cm_id, (void *)data->wr_id, (void *)data->wr_id, _recvMsgSize, _recvMr);
    if (req->type == MessageType::DELETE)
        processDelete(req);
    if (req->type == MessageType::GET)
        processGet(req);
    if (req->type == MessageType::PUT)
        processPut(req);
    delete[] request;
}

void RdmaSalEndpoint::processDelete(struct MessageHeader *req)
{
    rocksdb::Status s = _db->Delete(rocksdb::WriteOptions(), {(char *)req + MessageHeaderSize, req->keySize});
    void *sendBuf = nullptr;
    std::unique_lock<std::mutex> lock(_sendBuffersM);
    if (_sendBuffers.size() == 0)
        return;
    sendBuf = _sendBuffers.front();
    _sendBuffers.pop();
    lock.unlock();
    MessageHeader *response = (MessageHeader *)sendBuf;
    /*This id done to avoid else case*/
    response->type = MessageType::FAILURE;
    response->id = req->id;
    if (s.ok())
    {
        response->type = MessageType::SUCCESS;
    }
    rdma_post_send(_cm_id, sendBuf, sendBuf, MessageHeaderSize, _sendMr, 0);
}

void RdmaSalEndpoint::processGet(struct MessageHeader *req)
{
    char *sendBuf = nullptr;
    std::unique_lock<std::mutex> lock(_sendBuffersM);
    if (_sendBuffers.size() == 0)
    {
        return;
    }
    sendBuf = _sendBuffers.front();
    _sendBuffers.pop();
    lock.unlock();
    std::string value;
    rocksdb::Status s = _db->Get(rocksdb::ReadOptions(), {(char *)req + MessageHeaderSize, req->keySize}, &value);
    MessageHeader *response = (MessageHeader *)sendBuf;
    /*This id done to avoid else case*/
    response->type = MessageType::FAILURE;
    response->id = req->id;
    if (s.ok())
    {
        response->type = MessageType::SUCCESS;
        response->valueSize = value.size();
        memcpy(response + MessageHeaderSize, value.c_str(), value.size());
    }
    rdma_post_send(_cm_id, sendBuf, sendBuf, MessageHeaderSize + value.size(), _sendMr, 0);
}

void RdmaSalEndpoint::processPut(struct MessageHeader *req)
{
    rocksdb::Status s = _db->Put(rocksdb::WriteOptions(), {(char *)req + MessageHeaderSize, req->keySize},
                                 {(char *)req + MessageHeaderSize + req->keySize, req->valueSize});
    char *sendBuf = nullptr;
    std::unique_lock<std::mutex> lock(_sendBuffersM);
    if (_sendBuffers.size() == 0)
    {
        CPPLog::LOG_ERROR("No send Buffer");
        return;
    }
    sendBuf = _sendBuffers.front();
    _sendBuffers.pop();
    lock.unlock();
    MessageHeader *response = (MessageHeader *)sendBuf;
    /*This id done to avoid else case*/
    response->type = MessageType::FAILURE;
    response->id = req->id;
    if (s.ok())
        response->type = MessageType::FAILURE;
    rdma_post_send(_cm_id, sendBuf, sendBuf, MessageHeaderSize, _sendMr, 0);
}
