#include "RdmaServerEndpointGroup.hpp"

int RdmaServerEndpointGroup::CONN_STATE_INITIALIZED = 0;
int RdmaServerEndpointGroup::CONN_STATE_BINDED = 3;
int RdmaServerEndpointGroup::CONN_STATE_CONNECTED = 4;
int RdmaServerEndpointGroup::CONN_STATE_CLOSED = 5;

RdmaServerEndpointGroup::RdmaServerEndpointGroup(int sendQueueSize, int recvQueueSize, int compQueueSize,
                                                 int sendMsgSize, int recvMsgSize, std::string dbpath)
    : _sendQueueSize(sendQueueSize), _recvQueueSize(recvQueueSize), _compQueueSize(compQueueSize),
      _sendMsgSize(sendMsgSize), _recvMsgSize(recvMsgSize)
{
    _cmProcessor = new RdmaCmProcessor(this);
    _cm_id = _cmProcessor->createId();
    //_qpRepEndpointMap = new std::unordered_map<uint32_t, RdmaReplicationEndpoint *>();
    //_qpSalEndpointMap = new std::unordered_map<uint32_t, RdmaSalEndpoint *>();
    //_salEps = new std::vector<RdmaSalEndpoint *>();
    //_repEps = new std::vector<RdmaReplicationEndpoint *>();

    rocksdb::Options options;
    options.create_if_missing = true;
    // open a database with a name which corresponds to a file system directory
    rocksdb::Status status = rocksdb::DB::Open(options, dbpath, &_db);
    CPPLog::LOG_ALWAYS("Rocks started");
    if (!status.ok())
    {
        CPPLog::LOG_ERROR(status.ToString().c_str());
        exit(1);
    }
}
void RdmaServerEndpointGroup::setExecutor(Executor *ex)
{
    _executor = ex;
}
void RdmaServerEndpointGroup::startCmProcessor(bool newThread)
{
    _cmProcessor->start(newThread);
}

void RdmaServerEndpointGroup::bind(const char *ip, const char *port, int backlog)
{
    int ret;
    CPPLog::LOG_ALWAYS("RdmaServerEndpointGroup : Step 2 bind_addr");
    struct addrinfo *addr;
    ret = getaddrinfo(ip, port, NULL, &addr);
    if (ret)
    {
        CPPLog::LOG_ERROR("RdmaServerEndpointGroup : get_addr_info failed");
    }
    ret = rdma_bind_addr(_cm_id, addr->ai_addr);
    if (ret)
    {
        CPPLog::LOG_ERROR("RdmaServerEndpointGroup : rdma_bind_addr failed");
        return;
    }

    CPPLog::LOG_ALWAYS("RdmaServerEndpointGroup : rdma_bind_addr successful");
    ret = rdma_listen(_cm_id, backlog);
    if (ret)
    {
        CPPLog::LOG_ERROR("RdmaServerEndpointGroup : rdma_listen failed");
        return;
    }
}

struct ibv_cq *RdmaServerEndpointGroup::createSalCq(struct rdma_cm_id *id)
{
    if (_salCqProcessor == NULL)
    {
        CPPLog::LOG_ALWAYS("RdmaServerEndpointGroup : step 5 create salcq processor");
        _salCqProcessor = new RdmaSalCqProcessor(_executor, _cm_id->verbs, _compQueueSize);
        _salCqProcessor->start();
    }
    return _salCqProcessor->getCq();
}

struct ibv_cq *RdmaServerEndpointGroup::createRepCq(struct rdma_cm_id *id)
{
    if (_repCqProcessor == NULL)
    {
        CPPLog::LOG_ERROR("RdmaServerEndpointGroup : step 5 create repcq processor");
        _repCqProcessor = new RdmaRepCqProcessor(_executor, _cm_id->verbs, _compQueueSize);
        _repCqProcessor->start();
    }
    return _repCqProcessor->getCq();
}
void RdmaServerEndpointGroup::createEpCmEvent(struct rdma_cm_event *event)
{
    CPPLog::LOG_ALWAYS("RdmaServerEndpointGroup : step 4 Got Connect Request Server Endpoint");
    /*
     * create and add to vectors for replication and invalidation processing
     * connData is used to identify whether connection came for client or from other server;
     */

    const char *connData = reinterpret_cast<const char *>(event->param.conn.private_data);
    if (strcmp(connData, "sal") == 0)
    {
        // std::cout << "sal" << std::endl;
        RdmaSalEndpoint *endpoint = new RdmaSalEndpoint(event->id, createSalCq(event->id), _sendQueueSize, _recvQueueSize,
                                                        _sendMsgSize, _recvMsgSize, _db);
        event->id->context = (void *)endpoint;
        endpoint->createResources();
        std::unique_lock lock(_salMutex);
        _salEps.push_back(endpoint);
        // std::cout<<"qp num"<<event->id->qp->qp_num<<" "<<_qpSalEndpointMap->size();
        _qpSalEndpointMap.emplace(event->id->qp->qp_num, endpoint);
        //  std::cout<<_qpSalEndpointMap->size()<<std::endl;
    }
    else
    {
        RdmaReplicationEndpoint *endpoint = new RdmaReplicationEndpoint(event->id, createRepCq(event->id), _sendQueueSize, _recvQueueSize,
                                                                        _sendMsgSize, _recvMsgSize, _db);
        event->id->context = (void *)endpoint;
        endpoint->createResources();
        std::unique_lock lock(_repMutex);
        _repEps.push_back(endpoint);
        _qpRepEndpointMap.emplace(event->id->qp->qp_num, endpoint);
    }
    rdma_accept(event->id, NULL);
}

void RdmaServerEndpointGroup::processCmEvent(struct rdma_cm_event *event)
{
    std::ostringstream ss;
    ss << "RdmaServerEndpointGroup : event " << rdma_event_str(event->event);
    ss << " id " << event->id;
    CPPLog::LOG_ALWAYS(ss);
    /*
     * Connect request came on listener ie endpointgroup
     * in this listen_id contains cm_id of listener and id contains the new id of connection request
     */
    if (event->event == RDMA_CM_EVENT_CONNECT_REQUEST && event->listen_id != NULL)
    {
        createEpCmEvent(event);
    }
    else if (event->id != NULL && _cm_id == event->id)
    {
        if (event->event == RDMA_CM_EVENT_DISCONNECTED)
        {
            std::cout << "RdmaServerEndpointGroup : Disconnect Server Endpoint" << std::endl;
            close();
        }
        else
        {
            std::cout << "RdmaServerEndpointGroup : unknown Event for listener" << rdma_event_str(event->event) << std::endl;
        }
    }
    /*
     * Event came for a endpoint
     */
    else if (event->id != NULL && event->id->context != NULL)
    {
        /*
         * TIMEWAIT_EXIT happens only when client is already closed and we are not able to send
         * rdma_disconnect()
         * Since endpoint already closed with RDMA_CM_EVENT_DISCONNECTED
         * we need to return for this event
         */
        if (event->event == RDMA_CM_EVENT_TIMEWAIT_EXIT)
        {
            CPPLog::LOG_ALWAYS("Recieved TIMEWAIT_EXIT");
            return;
        }
        RdmaEndpoint *ep = ((RdmaEndpoint *)event->id->context);
        uint32_t qp  =  event->id->qp->qp_num;
        ep->processCmEvent(event);

        if (event->event == RDMA_CM_EVENT_DISCONNECTED)
        {
            {
                std::unique_lock lock(_repMutex);
                auto it = std::find(_repEps.begin(), _repEps.end(), (RdmaReplicationEndpoint *)ep);
                if (it != _repEps.end())
                {
                    _repEps.erase(it);
                    _qpRepEndpointMap.erase(qp);
                    delete ((RdmaReplicationEndpoint *)event->id->context);
                }
            }

            {
                std::unique_lock lock(_salMutex);
                auto it = std::find(_salEps.begin(), _salEps.end(), (RdmaSalEndpoint *)ep);
                if (it != _salEps.end())
                {
                    _salEps.erase(it);
                    _qpSalEndpointMap.erase(qp);
                    delete ((RdmaSalEndpoint *)event->id->context);
                }
            }
        }
    }
    else
    {
        std::ostringstream ss;
        ss << "RdmaServerEndpointGroup :  Not able to procces CM EVent";
        ss << rdma_event_str(event->event) << event->id << " ";
        ss << event->listen_id << std::endl;
        CPPLog::LOG_ERROR(ss);
    }
}

void RdmaServerEndpointGroup::close()
{
    _salCqProcessor->close();
    _repCqProcessor->close();
    _executor->stop();
    delete _salCqProcessor;
    delete _repCqProcessor;
    _cmProcessor->close();
    delete _cmProcessor;
    for (size_t i = 0; i < _salEps.size(); i++)
    {
        ((RdmaEndpoint *)_salEps[i])->close();
        delete _salEps[i];
    }
    for (size_t i = 0; i < _repEps.size(); i++)
    {
        ((RdmaEndpoint *)_repEps[i])->close();
        delete _repEps[i];
    }

    rdma_disconnect(_cm_id);
    rdma_destroy_id(_cm_id);
}