#include "TaskThread.hpp"

TaskThread::TaskThread(int id, int cpu, ConcurrentQueue *taskqueue, RdmaEndpointGroup *group)
    : _id(id), _group(group)
{
    _taskQueue = taskqueue;
    if (pthread_create(&thread, NULL, &TaskThread::run, this))
    {
        CPPLog::LOG_ERROR("pthread create has been failed while creating taskthread");
        exit(0);
    }
    cpu_set_t cpuset;
    CPU_ZERO(&cpuset);
    CPU_SET(cpu, &cpuset);
    std::ostringstream ss;
    ss<<"New Thread Setting CPU affinty "<<cpu;
    CPPLog::LOG_ALWAYS(ss);
    if (pthread_setaffinity_np(thread, sizeof(cpu_set_t), &cpuset) != 0)
    {
        CPPLog::LOG_ERROR("Error calling pthread_setaffinity_np ");
    }
}

TaskThread::TaskThread(int id, ConcurrentQueue *taskqueue, RdmaEndpointGroup *group)
    : _id(id), _group(group)
{
    _taskQueue = taskqueue;
    if (pthread_create(&thread, NULL, &TaskThread::run, this))
    {
        CPPLog::LOG_ERROR( "pthread create has been failed while creating taskthread");
        exit(0);
    }
    pthread_setname_np(thread,"TaskThread");
}

TaskThread::~TaskThread()
{
    CPPLog::LOG_INFO( "TaskThread Destructed");
    stop();
}

void TaskThread::stop()
{
    _stop = true;
    if (pthread_join(thread, NULL) == 0)
    {
        CPPLog::LOG_ERROR("pthread join failed");
    }
}

inline void *TaskThread::run(void *object)
{
    /*
    * This is because data from pthread_create is passed as void*
    */
    TaskThread *thread = reinterpret_cast<TaskThread *>(object);
    CPPLog::LOG_INFO("Running task thread");
    CPPLog::LOG_INFO( std::to_string( thread->_id).c_str());
    while (!thread->_stop)
    {
        struct ibv_wc *data = NULL;
        std::cout<<"Get start\n";
        data = thread->_taskQueue->try_pop();
        if (data != NULL)
        {
            std::cout<<"TaskThread:: got data";
            thread->processEvent(data);
            thread->_taskQueue->removeFromSet(data);
            delete data;
        }
    }
    return NULL;
}

void TaskThread::processEvent(struct ibv_wc *data)
{
    if (data == NULL || data->status != IBV_WC_SUCCESS)
    {
        std::ostringstream ss;
        ss<< "TaskThread : failed work completion : ";
        ss<<ibv_wc_status_str(data->status) << " on qp "<<  data->qp_num;
        CPPLog::LOG_ERROR(ss);
        return;
    }
    auto it = _group->_qpSalEndpointMap.find(data->qp_num);
    if (it == _group->_qpSalEndpointMap.end())
    {
        CPPLog::LOG_INFO("RdmaSal : endpoint not registered for qp");
        return;
    }
    // processSalCQEvent(data, it->second);
    switch (data->opcode)
    {
    case IBV_WC_SEND:
        it->second->processSendCompletion(data);
        break;
    case IBV_WC_RECV:
    {
        // it->second->processRecvCompletion(data);
        char *buffer = new char[data->byte_len];
        memcpy(buffer, (void *)data->wr_id, data->byte_len);
        rdma_post_recv(it->second->_cm_id, (void *)data->wr_id, (void *)data->wr_id,
                       it->second->_recvMsgSize, it->second->_recvMr);
        struct MessageHeader *req = (struct MessageHeader *)buffer;
        std::cout<<"TaskThread 1\n";
        switch (req->type)
        {
        case MessageType::GET:
            it->second->processGet(req);
            break;
        case MessageType::DELETE:
            replicateSalRequest(buffer, data->byte_len);
            it->second->processDelete(req);
            break;
        case MessageType::PUT:
            replicateSalRequest(buffer, data->byte_len);
            it->second->processPut(req);
            break;
        default:
            CPPLog::LOG_ERROR("SalRequest invalid req type");
            break;
        }
        delete[] buffer;
    }
    break;
    case IBV_WC_RDMA_WRITE:
        CPPLog::LOG_INFO("rdma write completion");
        break;
    case IBV_WC_RDMA_READ:
        CPPLog::LOG_INFO( "rdma read completion");
        break;
    default:
        CPPLog::LOG_INFO( "TaskThread default opcode : "); 
       // CPPLog::LOG_INFO(data->opcode);
        CPPLog::LOG_INFO("");
        break;
    }
}
void TaskThread::replicateSalRequest(char *req, uint32_t size)
{
    //send to follower to Replicate Request
//    for (auto i = _group->_repEps.begin(); i != _group->_repEps.end(); i++)
//    {
//        (*i)->sendMessage(req, size);
//    }
//    std::cout<<"size "<<_group->_repEps.size()<<" "<<_group->_repEps[0]<<"";
    std::cout<<"taskthread 2\n";
    for(auto i : _group->_repEps)
    {
        i->sendMessage(req,size);
    }
    
    MessageHeader *salReq = (MessageHeader *)req;
    char *buffer = new char[MessageHeaderSize + salReq->keySize];
    MessageHeader *invRequest = (MessageHeader *)(buffer);
    invRequest->type = MessageType::INVALIDATE;
    invRequest->id = salReq->id;
    invRequest->keySize = salReq->keySize;
    memcpy(buffer + MessageHeaderSize, salReq + MessageHeaderSize, salReq->keySize);
    //Send Invalidation to sal's
    for (auto i : _group->_salEps )
    {
        i->sendMessage(buffer, MessageHeaderSize + salReq->keySize);
    }
    delete[] buffer;
}