package hpdos_rdma_offloaded.service;

import java.io.IOException;
import java.net.SocketAddress;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;

import com.google.common.base.Stopwatch;
import com.ibm.darpc.DaRPCClientEndpoint;
import com.ibm.darpc.DaRPCClientGroup;
import com.ibm.darpc.DaRPCServerEndpoint;
import com.ibm.darpc.DaRPCServerEvent;
import com.ibm.darpc.DaRPCService;
import com.ibm.darpc.DaRPCStream;

import org.apache.log4j.Logger;
import org.rocksdb.RocksDBException;

import hpdos_rdma_offloaded.ExperimentStatistics;
import hpdos_rdma_offloaded.handler.NetworkHandler;
import hpdos_rdma_offloaded.handler.StorageHandler;
import hpdos_rdma_offloaded.lib.Packet;
import hpdos_rdma_offloaded.protocol.AckType;
import hpdos_rdma_offloaded.protocol.InvalidationRequest;
import hpdos_rdma_offloaded.protocol.InvalidationResponse;
import hpdos_rdma_offloaded.protocol.InvalidationRpcProtocol;
import hpdos_rdma_offloaded.protocol.Request;
import hpdos_rdma_offloaded.protocol.RequestType;
import hpdos_rdma_offloaded.protocol.Response;
import hpdos_rdma_offloaded.protocol.RpcProtocol;

public class MasterMetadataServerService extends RpcProtocol implements DaRPCService<Request, Response>{
    NetworkHandler networkHandler;
    StorageHandler storageHandler;
    ConcurrentHashMap<SocketAddress,DaRPCStream<InvalidationRequest, InvalidationResponse>> invalidationClientEpStreams;
    ConcurrentHashMap<SocketAddress,DaRPCClientEndpoint<InvalidationRequest, InvalidationResponse>> invalidationClientEps;
    DaRPCClientGroup<InvalidationRequest, InvalidationResponse> invalidationClientGroup;
    final static Logger logger = Logger.getLogger(MasterMetadataServerService.class);

    public MasterMetadataServerService(NetworkHandler networkHandler,StorageHandler storageHandler) throws Exception{
        logger.debug("New MetadataServerService Created");
        this.networkHandler = networkHandler;
        this.storageHandler  = storageHandler;
        this.invalidationClientEpStreams = new ConcurrentHashMap<>();
        this.invalidationClientEps = new ConcurrentHashMap<>();
        this.networkHandler.setClientStreams(this.invalidationClientEpStreams);
        InvalidationRpcProtocol rpcProtocol = new InvalidationRpcProtocol();
        invalidationClientGroup  =  DaRPCClientGroup.createClientGroup(rpcProtocol, 100, 0, 16, 16);  
    }

    @Override
    public void processServerEvent(DaRPCServerEvent<Request, Response> event) throws IOException {
        Request request = event.getReceiveMessage();
        Response response = event.getSendMessage();
        Packet packet = new Packet(request.getRequestType(), request.getKey(), request.getValue());
        logger.debug("Received "+request.getRequestType()+" Request for key: " + new String(request.getKey()));
        Stopwatch stopwatch = Stopwatch.createUnstarted();
        long startTime, endTime, timeTaken;
        if(request.getValue() != null)
            logger.debug(" value : "+ new String(request.getValue()));
        try{
            if (request.getRequestType() == RequestType.PUT) {
                startTime = System.nanoTime();
                this.storageHandler.create(request);
                endTime = System.nanoTime();
                timeTaken = (endTime-startTime) / 1000;
                ExperimentStatistics.collectStatistics(timeTaken, "localWriteTime");
                logger.debug("Time to write to local cache: " + timeTaken);
                // stopwatch.reset();
                // stopwatch.start();
                startTime = System.nanoTime();
                this.networkHandler.create(packet);
                endTime = System.nanoTime();
                timeTaken = (endTime-startTime) / 1000;
                logger.debug("Time to replicate: " + timeTaken );
                ExperimentStatistics.collectStatistics(timeTaken, "replicationTime");
                // stopwatch.stop();
                // logger.debug("Time to replicate: " + stopwatch.elapsed(TimeUnit.MICROSECONDS));
                // stopwatch.reset();
                startTime = System.nanoTime();
             //   this.networkHandler.sendInvalidateRequest(request);
                endTime = System.nanoTime();
                timeTaken = (endTime-startTime) / 1000;
                logger.debug("Time to invalidate: " + timeTaken);
                ExperimentStatistics.collectStatistics(timeTaken, "invalidationTime");
                ExperimentStatistics.collectStatistics(0, "write");
                // stopwatch.reset();
                // logger.debug("Time to invalidate: " + stopwatch.elapsed(TimeUnit.MICROSECONDS));
                // stopwatch.reset();
               //Yet to consume isDone 
                // var isDone  = this.networkHandler.consumeReplicationRequestFutures(replicationRequestFutures);

                response.setAck(AckType.SUCCESS);
                response.setKey(null);
                response.setValue(null);
            } 
            else if(request.getRequestType() == RequestType.GET) {
                stopwatch.start();
                byte[] value = this.storageHandler.read(request); 
                stopwatch.stop();
                logger.debug("Time to read: " + stopwatch.elapsed(TimeUnit.MICROSECONDS));
                stopwatch.reset();
                if(value == null)
                {
                    response.setAck(AckType.NOTFOUND);
                }
                else
                {
                    response.setAck(AckType.SUCCESS_WITH_VALUE);
                    response.setValue(value);
                }
            }
            else if(request.getRequestType() == RequestType.DELETE){
                stopwatch.start();
                this.storageHandler.delete(request);
                stopwatch.stop();
                logger.debug("Time to delete to local cache: " + stopwatch.elapsed(TimeUnit.MICROSECONDS));
                stopwatch.reset();
                stopwatch.start();
                this.networkHandler.delete(packet);
                stopwatch.stop();
                logger.debug("Time to delete: " + stopwatch.elapsed(TimeUnit.MICROSECONDS));
                stopwatch.reset();
                stopwatch.start();
               // this.networkHandler.sendInvalidateRequest(request);
                stopwatch.stop();
                logger.debug("Time to invalidate delete: " + stopwatch.elapsed(TimeUnit.MICROSECONDS));
                stopwatch.reset();
                // var replicationRequestFutures = this.networkHandler.sendReplicateRequest(request);
               //Yet to consume isDone 
                // var isDone  = this.networkHandler.consumeReplicationRequestFutures(replicationRequestFutures);
                response.setAck(AckType.SUCCESS);
            }
            else{
                response.setAck(AckType.INVALID_REQUEST);
            }
        }
        catch(RocksDBException | InterruptedException | ExecutionException ex){
                ex.printStackTrace();
                response.setAck(AckType.DBFAILED);
        }
        event.triggerResponse();
    }

    @Override
    public void open(DaRPCServerEndpoint<Request, Response> rpcClientEndpoint) {
        logger.debug("Received new connection  ");
        // try{
        //     System.out.println(rpcClientEndpoint.getDstAddr()+" "+rpcClientEndpoint.getSrcAddr());
        //     String clientIP = ((InetSocketAddress)rpcClientEndpoint.getDstAddr()).getHostName();
        //     InetSocketAddress clientSocketAddress = new InetSocketAddress(clientIP,1921);
        //     System.out.println("Creating Endpoint");
        //     var invalidationClientEp = invalidationClientGroup.createEndpoint();
        //     invalidationClientEp.connect(clientSocketAddress, 100);
        //     System.out.println("accepted");
        //     if(invalidationClientEp.isConnected()){
        //         invalidationClientEpStreams.put(rpcClientEndpoint.getDstAddr(),invalidationClientEp.createStream());
        //         invalidationClientEps.put(rpcClientEndpoint.getDstAddr(),invalidationClientEp);
        //     }
        //     else{
        //         System.out.println("Not able to Connect to client for invalidation");
        //     }
        // }
        // catch(Exception e){
        //     e.printStackTrace();
        // }
    }

    @Override
    public void close(DaRPCServerEndpoint<Request, Response> rpcClientEndpoint) {
        try {
            SocketAddress socketAddress = rpcClientEndpoint.getDstAddr();
            String ipAddress = socketAddress.toString();
            int startIndex = ipAddress.indexOf("/") + 1;
            int endIndex = ipAddress.indexOf(":");

            ipAddress = ipAddress.substring(startIndex, endIndex);

            logger.debug("Closing Connection for ip address: " + ipAddress );
            NetworkHandler.invalidationStreams.remove(ipAddress);
            ExperimentStatistics.printStatistics();
        } catch (IOException e) {
            e.printStackTrace();
        }
        
    }
    
}
